@@ -30,6 +30,7 @@
 #include "action.h"
 #include "buffer.h"
 #include "command.h"
+#include "db.h"
 #include "err.h"
 #include "output.h"
 #include "parser.h"
@@ -37,7 +38,7 @@
 #include "stream.h"
 
 
-static void go(SQLHDBC conn, buffer *sqlbuf, char action, stream *stream)
+static void go(buffer *sqlbuf, char action, parsed_line *params, stream *stream)
 {
 	results *res = NULL;
 
@@ -46,16 +47,16 @@ static void go(SQLHDBC conn, buffer *sqlbuf, char action, stream *stream)
 		// do nothing
 		break;
 	case BUFFER_COMMAND:
-		res = run_command(conn, sqlbuf);
+		res = run_command(sqlbuf);
 		break;
 	case BUFFER_SQL:
-		res = execute_query(conn, sqlbuf->buf, sqlbuf->next);
+		res = execute_query(sqlbuf->buf, sqlbuf->next, params);
 		break;
 	}
 
 	if(res) {
 		output_results(res, action, stream);
-		results_free(res);
+		res_free(res);
 	}
 }
 
@@ -107,11 +108,10 @@ static void print(buffer *sqlbuf, stream *stream)
 	stream_putwc(stream, L'\n');
 }
 
-void run_action(SQLHDBC conn, buffer *sqlbuf, char action, char *paramstring)
+void run_action(buffer *sqlbuf, char action, char *paramstring)
 {
 	parsed_line *l;
-	int nchunks;
-	char *pipeline, *p;
+	char *pipeline;
 	FILE *f;
 	stream *stream;
 	int m;
@@ -120,18 +120,14 @@ void run_action(SQLHDBC conn, buffer *sqlbuf, char action, char *paramstring)
 	m = 0;
 
 	l = parse_string(paramstring);
-	nchunks = l->nchunks;
 
-	if(nchunks) {
-		p = l->chunks[nchunks - 1];
-		if(*p == '>') {
-			if(!(pipeline = malloc(strlen(p) + 5))) err_system();
+	if(l->pipeline) {
+		if(*l->pipeline == '>') {
+			if(!(pipeline = malloc(strlen(l->pipeline) + 5))) err_system();
 			m = 1;
-			sprintf(pipeline, "cat %s", p);
-			nchunks--;
-		} else if(*p == '|') {
-			pipeline = p + 1;
-			nchunks--;
+			sprintf(pipeline, "cat %s", l->pipeline);
+		} else if(*l->pipeline == '|') {
+			pipeline = l->pipeline + 1;
 		}
 	}
 
@@ -167,13 +163,13 @@ void run_action(SQLHDBC conn, buffer *sqlbuf, char action, char *paramstring)
 		// TODO: save to named buffer
 		break;
 	default:
-		go(conn, sqlbuf, action, stream);
+		go(sqlbuf, action, l, stream);
 		break;
 	}
 
 
 	stream_reset(stream);
-	if(pipeline) pclose(stream->f);
+	if(pipeline) pclose(f);
 	free(stream);
 
 	free_parsed_line(l);
