@@ -44,34 +44,48 @@ static results *get_help(const char *topic)
 	if(!strcmp(topic, "intro")) text = HELP_INTRO;
 	else text = HELP_NOTFOUND;
 
-	res = results_single_alloc();
-	results_set_cols(res, 1, topic);
-	results_add_row(res, text);
+	res = res_alloc();
+	res_set_cols(res, 1, topic);
+	res_add_row(res, text);
 
 	return res;
 }
 
 static results *get_copying()
 {
-	results *res = results_single_alloc();
-	results_set_cols(res, 1, _("COPYING"));
-	results_add_row(res, GPL_COPYING);
+	results *res = res_alloc();
+	res_set_cols(res, 1, _("COPYING"));
+	res_add_row(res, GPL_COPYING);
 	return res;
 }
 
 static results *get_warranty()
 {
-	results *res = results_single_alloc();
-	results_set_cols(res, 1, _("WARRANTY"));
-	results_add_row(res, GPL_WARRANTY);
+	results *res = res_alloc();
+	res_set_cols(res, 1, _("WARRANTY"));
+	res_add_row(res, GPL_WARRANTY);
 	return res;
 }
 
+static results *autocommit(const char *state)
+{
+	int change;
+
+	if(state) {
+		if(!strcmp(state, "Off") ||
+		   !strcmp(state, "off") ||
+		   !strcmp(state, "0")) change = -1;
+		else change = 1;
+	} else change = 0;
+
+	return db_autocommit(change);
+}
+
 static results *set(const char *name, const char *value)
 {
-	results *res = results_single_alloc();
+	results *res = res_alloc();
 
-	results_set_cols(res, 2, _("name"), _("value"));
+	res_set_cols(res, 2, _("name"), _("value"));
 
 	if(name) {
 		char *prefixed_name;
@@ -82,21 +96,18 @@ static results *set(const char *name, const char *value)
 			if(setenv(prefixed_name, value, 1) == -1) {
 				char buf[64];
 				strerror_r(errno, buf, 64);
-				results_set_warnings(res, 1, buf);
+				res_add_warning(res, buf);
 			}
 		} else {
 			value = getenv(prefixed_name);
-			if(!value) results_set_warnings(res, 1, _("Variable not set"));
+			if(!value) res_add_warning(res, _("Variable not set"));
 		}
 
-		results_add_row(res, name, value);
+		res_add_row(res, name, value);
 
 		free(prefixed_name);
 	} else {
 		char **v, *name, *value;
-		row **rp;
-
-		rp = &(res->sets->rows);
 
 		for(v = environ; *v; v++) {
 			if(!strncmp(*v, "DBSH_", 5)) {
@@ -109,13 +120,7 @@ static results *set(const char *name, const char *value)
 
 				if(*value) {
 					*value++ = 0;
-
-					*rp = results_row_alloc(2);
-					(*rp)->data[0] = strdup2wcs(name);
-					(*rp)->data[1] = strdup2wcs(value);
-
-					res->sets->nrows++;
-					rp = &(*rp)->next;
+					res_add_row(res, name, value);
 				}
 
 				free(name);
@@ -142,43 +147,58 @@ static void unset(const char *name)
 	free(prefixed_name);
 }
 
-results *run_command(SQLHDBC conn, buffer *buf)
+#define SYNTAX(p) printf(_("Syntax: %s %s\n"), c, p)
+
+results *run_command(buffer *buf)
 {
 	parsed_line *l;
 	results *res = 0;
+	char *c, *p1, *p2;
 
 	l = parse_buffer(buf);
 	if(l->nchunks < 1) return 0;
 
+	c  = l->chunks[0] + 1;
+	p1 = l->nchunks > 1 ? l->chunks[1] : 0;
+	p2 = l->nchunks > 2 ? l->chunks[2] : 0;
+
 	// Help commands
-	if(l->chunks[0][1] == 'h') res = get_help(l->nchunks > 1 ? l->chunks[1] : "intro");
-	else if(!strncmp(l->chunks[0] + 1, "cop", 3)) res = get_copying();
-	else if(!strncmp(l->chunks[0] + 1, "war", 3)) res = get_warranty();
+	if(*c == 'h') res = get_help(p1 ? p1 : "intro");
+	else if(!strncmp(c, "cop", 3)) res = get_copying();
+	else if(!strncmp(c, "war", 3)) res = get_warranty();
 
 	// Catalog commands
-	else if(!strncmp(l->chunks[0] + 1, "cat", 3)) {
-		res = get_tables(conn, SQL_ALL_CATALOGS, "", "");
-	} else if(!strncmp(l->chunks[0] + 1, "sch", 3)) {
-		res = db_list_schemas(conn, l->nchunks > 1 ? l->chunks[1] : 0);
-	} else if(!strncmp(l->chunks[0] + 1, "tab", 3)) {
-		res = db_list_tables(conn, l->nchunks > 1 ? l->chunks[1] : 0);
-	} else if(!strncmp(l->chunks[0] + 1, "col", 3)) {
-		if(l->nchunks > 1) res = db_list_columns(conn, l->chunks[1]);
-		else printf(_("Syntax: columns <table>\n"));
+	else if(!strncmp(c, "cat", 3)) {
+		res = get_tables(SQL_ALL_CATALOGS, "", "");
+	} else if(!strncmp(c, "sch", 3)) {
+		res = get_tables("", SQL_ALL_SCHEMAS, "");
+	} else if(!strncmp(c, "tab", 3)) {
+		res = db_list_tables(p1);
+	} else if(!strncmp(c, "col", 3)) {
+		if(p1) res = db_list_columns(p1);
+		else SYNTAX(_("<table>"));
+	}
+
+	// Transaction commands
+	else if(!strcmp(c, "autocommit")) {
+		res = autocommit(p1);
+	} else if(!strcmp(c, "commit")) {
+		res = db_endtran(1);
+	} else if(!strcmp(c, "rollback")) {
+		res = db_endtran(0);
 	}
 
 	// Other commands
-	else if(!strcmp(l->chunks[0] + 1, "set")) {
-		res = set(l->nchunks > 1 ? l->chunks[1] : 0,
-			  l->nchunks > 2 ? l->chunks[2] : 0);
-	} else if(!strcmp(l->chunks[0] + 1, "unset")) {
-		if(l->nchunks > 1) unset(l->chunks[1]);
-		else printf(_("Syntax: unset <variable>"));
-	} else if(!strncmp(l->chunks[0] + 1, "inf", 3)) {
-		res = db_conn_details(conn);
+	else if(!strcmp(c, "set")) {
+		res = set(p1, p2);
+	} else if(!strcmp(c, "unset")) {
+		if(p1) unset(p1);
+		else SYNTAX(_("<variable>"));
+	} else if(!strncmp(c, "inf", 3)) {
+		res = db_conn_details();
 	}
 
-	else printf(_("Unrecognised command: %s\n"), l->chunks[0] + 1);
+	else printf(_("Unrecognised command: %s\n"), c);
 
 	free_parsed_line(l);
 
