@@ -25,23 +25,25 @@
 #include "buffer.h"
 #include "db.h"
 #include "err.h"
+#include "parser.h"
 #include "results.h"
 
 #define report_error(t, h, r, f) _report_error(t, h, r, f, __FILE__, __LINE__)
 
 extern const char *dsn, *user, *pass;
-extern SQLHDBC conn;
+SQLHDBC conn;
 
 SQLHSTMT *current_statement;
 pthread_mutex_t cs_lock = PTHREAD_MUTEX_INITIALIZER;
 
-static int list_dsns(SQLHENV, SQLUSMALLINT);
-static void time_taken(struct timeval *);
-static results *fetch_results(SQLHSTMT, struct timeval);
-static resultset *fetch_resultset(SQLHSTMT, buffer *);
-static row *fetch_row(SQLHSTMT, buffer *, int);
-static char *get_current_catalog(SQLHDBC);
-static void parse_catalog_spec(SQLHDBC, char *, char **, char **);
+
+static void set_current_statement(SQLHSTMT *);
+static void fetch_warnings(results *, SQLSMALLINT, SQLHANDLE);
+static void fetch_results(results *, SQLHSTMT);
+static void fetch_resultset(results *, SQLHSTMT, buffer *);
+static int fetch_row(results *, SQLHSTMT, buffer *);
+static char *get_current_catalog();
+static void parse_catalog_spec(char *, char **, char **);
 static void parse_qualified_table(char *, char **, char **);
 
 static void _report_error(SQLSMALLINT type, SQLHANDLE handle, SQLRETURN r,
@@ -99,12 +101,12 @@ static SQLHENV alloc_env()
 	if(!env) {
 
 		r = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &env);
-		if(!SQL_SUCCEEDED(r)) err_fatal(_("Failed to allocate environment handle\n"));
+		if(!SQL_SUCCEEDED(r)) err_fatal(_("Failed to allocate environment handle"));
 
 		r = SQLSetEnvAttr(env, SQL_ATTR_ODBC_VERSION, (SQLPOINTER) SQL_OV_ODBC3, 0);
 		if(!SQL_SUCCEEDED(r)) {
 			SQLFreeHandle(SQL_HANDLE_ENV, env);
-			err_fatal(_("Failed to set ODBC version to 3\n"));
+			err_fatal(_("Failed to set ODBC version to 3"));
 		}
 
 	}
@@ -112,53 +114,66 @@ static SQLHENV alloc_env()
 	return env;
 }
 
-void list_all_dsns()
+results *db_drivers_and_dsns()
 {
-	SQLHENV env = alloc_env();
+	SQLHENV env;
+	results *res;
+	SQLUSMALLINT dir;
+	SQLCHAR buf1[256], buf2[256];
 
-	puts(_("User Data Sources\n-----------------"));
-	if(!list_dsns(env, SQL_FETCH_FIRST_USER)) puts(_("(none)"));
-	puts(_("\nSystem Data Sources\n-------------------"));
-	if(!list_dsns(env, SQL_FETCH_FIRST_SYSTEM)) puts(_("(none)"));
-}
+	env = alloc_env();
 
-static int list_dsns(SQLHENV env, SQLUSMALLINT dir)
-{
-	SQLCHAR dsn_buf[256], des_buf[256];
-	SQLSMALLINT dsn_len, des_len;
-	int n = 0;
+	res = res_alloc();
+	res_set_cols(res, 1, _("Driver"));
+	while(SQL_SUCCEEDED(SQLDrivers(env, SQL_FETCH_NEXT,
+				       buf1, sizeof(buf1), 0, 0, 0, 0)))
+		res_add_row(res, buf1);
 
-	while(SQLDataSources(env, dir,
-			     dsn_buf, sizeof(dsn_buf), &dsn_len,
-			     des_buf, sizeof(des_buf), &des_len) == SQL_SUCCESS) {
+	res_add_set(res);
+	res_set_cols(res, 3, _("DSN"), _("Type"), _("Description"));
 
-		n++;
-		printf("%s (%s)\n", dsn_buf, des_buf);
+	dir = SQL_FETCH_FIRST_USER;
+	while(SQL_SUCCEEDED(SQLDataSources(env, dir,
+					   buf1, sizeof(buf1), 0,
+					   buf2, sizeof(buf2), 0))) {
+		res_add_row(res, buf1, _("user"), buf1);
 		dir = SQL_FETCH_NEXT;
 	}
 
-	return n;
+	dir = SQL_FETCH_FIRST_SYSTEM;
+	while(SQL_SUCCEEDED(SQLDataSources(env, dir,
+					   buf1, sizeof(buf1), 0,
+					   buf2, sizeof(buf2), 0))) {
+		res_add_row(res, buf1, _("system"), buf1);
+		dir = SQL_FETCH_NEXT;
+	}
+
+	return res;
 }
 
-SQLHDBC db_connect()
+static SQLHDBC connect()
 {
 	SQLHENV env;
 	SQLHDBC conn;
+	SQLSMALLINT ldsn, luser, lpass;
 	SQLRETURN r;
 
 	env = alloc_env();
 
 	r = SQLAllocHandle(SQL_HANDLE_DBC, env, &conn);
-	if(!SQL_SUCCEEDED(r)) err_fatal(_("Failed to allocate connection handle\n"));
+	if(!SQL_SUCCEEDED(r)) err_fatal(_("Failed to allocate connection handle"));
 
+	ldsn =  dsn  ? SQL_NTS : 0;
+	luser = user ? SQL_NTS : 0;
+	lpass = pass ? SQL_NTS : 0;
 
 	if(!strchr(dsn, '=') || user) {
 		r = SQLConnect(conn,
-			       (SQLCHAR *) dsn, SQL_NTS,
-			       (SQLCHAR *) user, SQL_NTS,
-			       (SQLCHAR *) pass, SQL_NTS);
+			       (SQLCHAR *) dsn, ldsn,
+			       (SQLCHAR *) user, luser,
+			       (SQLCHAR *) pass, lpass);
 	} else {
-		r = SQLDriverConnect(conn, 0, (SQLCHAR *) dsn, SQL_NTS,
+		r = SQLDriverConnect(conn, 0, (SQLCHAR *) dsn, ldsn,
 				     0, 0, 0, SQL_DRIVER_NOPROMPT);
 	}
 
@@ -174,18 +189,31 @@ SQLHDBC db_connect()
 	return conn;
 }
 
+int db_connect()
+{
+	conn = connect();
+	return conn ? 1 : 0;
+}
+
 void db_reconnect()
 {
 	SQLHDBC newconn;
 
-	if((newconn = db_connect(dsn, user, pass))) {
+	if((newconn = connect())) {
 		SQLDisconnect(conn);
 		SQLFreeHandle(SQL_HANDLE_DBC, conn);
 		conn = newconn;
 	}
 }
 
-SQLSMALLINT db_info(SQLHDBC conn, SQLUSMALLINT type, char *buf, int len)
+void db_close()
+{
+	SQLDisconnect(conn);
+	SQLFreeHandle(SQL_HANDLE_DBC, conn);
+	SQLFreeHandle(SQL_HANDLE_ENV, alloc_env());
+}
+
+SQLSMALLINT db_info(SQLUSMALLINT type, char *buf, int len)
 {
 	SQLRETURN r;
 	SQLSMALLINT l;
@@ -201,7 +229,7 @@ SQLSMALLINT db_info(SQLHDBC conn, SQLUSMALLINT type, char *buf, int len)
 	return l;
 }
 
-SQLINTEGER db_conn_attr(SQLHDBC conn, SQLINTEGER attr, char *buf, int len)
+SQLINTEGER db_conn_attr(SQLINTEGER attr, char *buf, int len)
 {
 	SQLRETURN r;
 	SQLINTEGER l;
@@ -217,18 +245,18 @@ SQLINTEGER db_conn_attr(SQLHDBC conn, SQLINTEGER attr, char *buf, int len)
 	return l;
 }
 
-#define ADD_INFO(t, n) db_info(conn, t, buf, 256); results_add_row(res, n, buf)
+#define ADD_INFO(t, n) db_info(t, buf, 256); res_add_row(res, n, buf)
 
-results *db_conn_details(SQLHDBC conn)
+results *db_conn_details()
 {
 	results *res;
 	char buf[256];
 	SQLUINTEGER i;
 	const char *s;
 
-	res = results_single_alloc();
+	res = res_alloc();
 
-	results_set_cols(res, 2, _("name"), _("value"));
+	res_set_cols(res, 2, _("name"), _("value"));
 
 	// Connection info
 	ADD_INFO(SQL_DATA_SOURCE_NAME, _("DSN"));
@@ -256,7 +284,7 @@ results *db_conn_details(SQLHDBC conn)
 	default:
 		s = _("(unknown)");
 	}
-	results_add_row(res, _("ODBC compliance"), s);
+	res_add_row(res, _("ODBC compliance"), s);
 
 	SQLGetInfo(conn, SQL_SQL_CONFORMANCE, &i, 0, 0);
 	switch(i) {
@@ -275,7 +303,7 @@ results *db_conn_details(SQLHDBC conn)
 	default:
 		s = _("(unknown)");
 	}
-	results_add_row(res, _("SQL-92 compliance"), s);
+	res_add_row(res, _("SQL-92 compliance"), s);
 
 	// Driver Manager info
 	ADD_INFO(SQL_DM_VER,         _("Driver Manager version"));
@@ -288,7 +316,6 @@ results *db_conn_details(SQLHDBC conn)
 	ADD_INFO(SQL_TABLE_TERM,     _("Table term"));
 	ADD_INFO(SQL_PROCEDURE_TERM, _("Procedure term"));
 
-
 	return res;
 }
 
@@ -303,11 +330,12 @@ int db_supports_catalogs(SQLHDBC conn)
 	return (buf[0] == 'Y');
 }
 
-results *execute_query(SQLHDBC conn, const char *buf, int buflen)
+results *execute_query(const char *buf, int buflen, parsed_line *params)
 {
 	SQLHSTMT st;
+	results *res;
+	int i, l;
 	SQLRETURN r;
-	struct timeval taken;
 
 	r = SQLAllocHandle(SQL_HANDLE_STMT, conn, &st);
 	if(!SQL_SUCCEEDED(r)) {
@@ -315,110 +343,147 @@ results *execute_query(SQLHDBC conn, const char *buf, int buflen)
 		return 0;
 	}
 
-	pthread_mutex_lock(&cs_lock);
-	current_statement = &st;
-	pthread_mutex_unlock(&cs_lock);
+	set_current_statement(&st);
 
-	gettimeofday(&taken, 0);
+	res = res_alloc();
+	res_start_timer(res);
 
-	if(SQLExecDirect(st, (SQLCHAR *) buf, buflen) == SQL_ERROR) {
-		report_error(SQL_HANDLE_STMT, st, r, _("Failed to execute statement"));
+	r = SQLPrepare(st, (SQLCHAR *) buf, buflen);
+	if(!SQL_SUCCEEDED(r)) {
+		report_error(SQL_HANDLE_STMT, st, r, _("Failed to prepare statement"));
 		SQLFreeHandle(SQL_HANDLE_STMT, st);
+		res_free(res);
 		return 0;
+	} else if(r == SQL_SUCCESS_WITH_INFO) {
+		fetch_warnings(res, SQL_HANDLE_STMT, st);
 	}
 
-	time_taken(&taken);
+	for(i = 0; i < params->nchunks; i++) {
+		l = strlen(params->chunks[i]);
 
-	return fetch_results(st, taken);
-}
-
-static results *fetch_results(SQLHSTMT st, struct timeval time_taken)
-{
-	results *res;
-	resultset **sp;
-	buffer *buf;
-	SQLINTEGER i;
-	SQLSMALLINT reqlen;
-	SQLRETURN r;
-
-	res = results_alloc();
-	res->time_taken = time_taken;
-
-	buf = buffer_alloc(1024);
-
-	SQLGetDiagField(SQL_HANDLE_STMT, st, 0, SQL_DIAG_NUMBER, &(res->nwarnings), 0, 0);
-	if(res->nwarnings) {
-		if(!(res->warnings = calloc(res->nwarnings, sizeof(char *)))) err_system();
-		for(i = 0; i < res->nwarnings; i++) {
-			SQLGetDiagField(SQL_HANDLE_STMT, st, i + 1,
-					SQL_DIAG_MESSAGE_TEXT, buf->buf, buf->len, &reqlen);
-
-			if(reqlen + 1 > buf->len) {
-				buffer_realloc(buf, reqlen + 1);
-				SQLGetDiagField(SQL_HANDLE_STMT, st, i + 1,
-						SQL_DIAG_MESSAGE_TEXT, buf->buf, buf->len, 0);
-			}
-
-			if(!(res->warnings[i] = strdup2wcs(buf->buf))) err_system();
+		r = SQLBindParameter(st, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR,
+				     SQL_CHAR, l, 0, params->chunks[i], l, 0);
+		if(!SQL_SUCCEEDED(r)) {
+			report_error(SQL_HANDLE_STMT, st, r, _("Failed to bind parameter"));
+			SQLFreeHandle(SQL_HANDLE_STMT, st);
+			res_free(res);
+			return 0;
+		} else if(r == SQL_SUCCESS_WITH_INFO) {
+			fetch_warnings(res, SQL_HANDLE_STMT, st);
 		}
 	}
 
-	sp = &res->sets;
+	r = SQLExecute(st);
+	if(!SQL_SUCCEEDED(r) && r != SQL_NO_DATA) {
+		report_error(SQL_HANDLE_STMT, st, r, _("Failed to execute statement"));
+		SQLFreeHandle(SQL_HANDLE_STMT, st);
+		res_free(res);
+		return 0;
+	} else if(r == SQL_SUCCESS_WITH_INFO) {
+		fetch_warnings(res, SQL_HANDLE_STMT, st);
+	}
 
-	do {
-		*sp = fetch_resultset(st, buf);
-		sp = &(*sp)->next;
-		r = SQLMoreResults(st);
-	} while(SQL_SUCCEEDED(r));
-
-	buffer_free(buf);
-
-	pthread_mutex_lock(&cs_lock);
-	current_statement = 0;
-	pthread_mutex_unlock(&cs_lock);
-
-	SQLFreeHandle(SQL_HANDLE_STMT, st);
+	fetch_results(res, st);
+	res_stop_timer(res);
 
 	return res;
 }
 
-static resultset *fetch_resultset(SQLHSTMT st, buffer *buf)
+static void set_current_statement(SQLHSTMT *stp)
 {
-	resultset *res;
-	SQLRETURN r;
-	SQLSMALLINT i;
+	pthread_mutex_lock(&cs_lock);
+	current_statement = stp;
+	pthread_mutex_unlock(&cs_lock);
+}
+
+static void fetch_warnings(results *res, SQLSMALLINT type, SQLHANDLE h)
+{
+	SQLINTEGER n, i;
+	buffer *buf;
 	SQLSMALLINT reqlen;
-	row **rowp;
 
-	res = resultset_alloc();
+	SQLGetDiagField(SQL_HANDLE_STMT, h, 0, SQL_DIAG_NUMBER, &n, 0, 0);
+	if(!n) return;
 
-	r = SQLNumResultCols(st, &(res->ncols));
-	if(!SQL_SUCCEEDED(r)) {
-		report_error(SQL_HANDLE_STMT, st, r, _("Failed to retrieve number of columns"));
-		resultset_free(res);
-		return 0;
-	}
+	buf = buffer_alloc(1024);
+
+	for(i = 0; i < n; i++) {
+		SQLGetDiagField(SQL_HANDLE_STMT, h, i + 1, SQL_DIAG_MESSAGE_TEXT,
+				buf->buf, buf->len, &reqlen);
+
+		if(reqlen + 1 > buf->len) {
+			buffer_realloc(buf, reqlen + 1);
+			SQLGetDiagField(SQL_HANDLE_STMT, h, i + 1, SQL_DIAG_MESSAGE_TEXT,
+					buf->buf, buf->len, 0);
 
-	if(!res->ncols) {  // non-SELECT
-		r = SQLRowCount(st, &(res->nrows));
-		if(!SQL_SUCCEEDED(r)) {
-			report_error(SQL_HANDLE_STMT, st, r, _("Failed to retrieve rows affected"));
-			resultset_free(res);
-			return 0;
 		}
 
-		return res;
+		res_add_warning(res, buf->buf);
 	}
 
-	if(!(res->cols = calloc(res->ncols, sizeof(char *)))) err_system();
+	buffer_free(buf);
+}
 
-	for(i = 0; i < res->ncols; i++) {
+void fetch_results(results *res, SQLHSTMT st)
+{
+	buffer *buf;
+	SQLRETURN r;
+
+	buf = buffer_alloc(1024);
+
+	for(;;) {
+		fetch_resultset(res, st, buf);
+
+		r = SQLMoreResults(st);
+
+		if(r == SQL_NO_DATA) {
+			break;
+		} else if(!SQL_SUCCEEDED(r)) {
+			fetch_warnings(res, SQL_HANDLE_STMT, st);
+			break;
+		} else if(r == SQL_SUCCESS_WITH_INFO) {
+			fetch_warnings(res, SQL_HANDLE_STMT, st);
+		}
+
+		res_add_set(res);
+	}
+
+	buffer_free(buf);
+
+	set_current_statement(0);
+	SQLFreeHandle(SQL_HANDLE_STMT, st);
+}
+
+void fetch_resultset(results *res, SQLHSTMT st, buffer *buf)
+{
+	SQLSMALLINT ncols, i, reqlen;
+	SQLLEN nrows;
+	SQLRETURN r;
+
+	r = SQLNumResultCols(st, &ncols);
+	if(!SQL_SUCCEEDED(r)) {
+		report_error(SQL_HANDLE_STMT, st, r, _("Failed to retrieve number of columns"));
+		return;
+	}
+	res_set_ncols(res, ncols);
+
+	if(!ncols) {  // non-SELECT
+		r = SQLRowCount(st, &nrows);
+		if(!SQL_SUCCEEDED(r)) {
+			report_error(SQL_HANDLE_STMT, st, r, _("Failed to retrieve rows affected"));
+			return;
+		}
+
+		res_set_nrows(res, nrows);
+		return;
+	}
+
+	for(i = 0; i < ncols; i++) {
 		SQLSMALLINT type;
 		SQLUINTEGER size;
 		SQLSMALLINT digits;
 		SQLSMALLINT nullable;
 
-		// TODO: is there a way to just get the name?
 		r = SQLDescribeCol(st, i + 1, (SQLCHAR *) buf->buf, buf->len, &reqlen,
 				   &type, &size, &digits, &nullable);
 
@@ -430,54 +495,55 @@ static resultset *fetch_resultset(SQLHSTMT st, buffer *buf)
 
 		if(!SQL_SUCCEEDED(r)) {
 			report_error(SQL_HANDLE_STMT, st, r, _("Failed to retrieve column data"));
-			resultset_free(res);
-			return 0;
+			return;
 		}
 
-		if(!(res->cols[i] = strdup2wcs(buf->buf))) err_system();
+		res_set_col(res, i, buf->buf);
 	}
 
-	for(rowp = &(res->rows);;) {
-		*rowp = fetch_row(st, buf, res->ncols);
-		if(!*rowp) break;
-		rowp = &(*rowp)->next;
-		res->nrows++;
-	}
 
-	return res;
+	while(fetch_row(res, st, buf));
 }
 
-static row *fetch_row(SQLHSTMT st, buffer *buf, int ncols)
+static int fetch_row(results *res, SQLHSTMT st, buffer *buf)
 {
-	row *row;
 	SQLRETURN r;
 	SQLSMALLINT i;
-	SQLINTEGER reqlen;
+	SQLINTEGER reqlen, offset;
+
 
 	r = SQLFetch(st);
 	if(!SQL_SUCCEEDED(r)) return 0;
 
-	row = results_row_alloc(ncols);
+	res_new_row(res);
 
-	for(i = 0; i < ncols; i++) {
-		r = SQLGetData(st, i + 1, SQL_C_CHAR, buf->buf, buf->len, &reqlen);
+	for(i = 0; i < res_get_ncols(res); i++) {
+		offset = 0;
 
-		if(reqlen + 1 > buf->len) {
-			buffer_realloc(buf, reqlen + 1);
-			r = SQLGetData(st, i + 1, SQL_C_CHAR, buf->buf, buf->len, 0);
+		for(;;) {
+			r = SQLGetData(st, i + 1, SQL_C_CHAR,
+				       buf->buf + offset, buf->len - offset,
+				       &reqlen);
+
+			if(r == SQL_NO_DATA) break;
+			else if(!SQL_SUCCEEDED(r)) {
+				report_error(SQL_HANDLE_STMT, st, r, _("Failed to fetch row"));
+				return 0;
+			}
+
+			if(reqlen == SQL_NULL_DATA) break;
+			else if(reqlen == SQL_NO_TOTAL) reqlen = buf->len + 1024;  // guess
+
+			if(reqlen > 0 && reqlen + 1 > buf->len) {
+				offset = buf->len - 1;
+				buffer_realloc(buf, reqlen + 1);
+			} else break;
 		}
 
-		if(!SQL_SUCCEEDED(r)) {
-			report_error(SQL_HANDLE_STMT, st, r, _("Failed to fetch row"));
-			results_row_free(row, ncols);
-			return 0;
-		}
-
-		if(reqlen != SQL_NULL_DATA &&
-		   !(row->data[i] = strdup2wcs(buf->buf))) err_system();
+		if(reqlen != SQL_NULL_DATA) res_set_value(res, i, buf->buf);
 	}
 
-	return row;
+	return 1;
 }
 
 void db_cancel_query()
@@ -494,13 +560,12 @@ void db_cancel_query()
 	pthread_mutex_unlock(&cs_lock);
 }
 
-results *get_tables(SQLHDBC conn, const char *catalog,
+results *get_tables(const char *catalog,
 		       const char *schema, const char *table)
 {
 	SQLHSTMT st;
 	SQLRETURN r;
-	struct timeval taken;
-
+	results *res;
 
 	r = SQLAllocHandle(SQL_HANDLE_STMT, conn, &st);
 	if(!SQL_SUCCEEDED(r)) {
@@ -508,29 +573,39 @@ results *get_tables(SQLHDBC conn, const char *catalog,
 		return 0;
 	}
 
-	gettimeofday(&taken, 0);
+	set_current_statement(&st);
+
+	res = res_alloc();
+	res_start_timer(res);
+
 	current_statement = st;
 	r = SQLTables(st,
 		      (SQLCHAR *) catalog, SQL_NTS,
 		      (SQLCHAR *) schema, SQL_NTS,
 		      (SQLCHAR *) table, SQL_NTS,
 		      (SQLCHAR *) 0, 0);
-	time_taken(&taken);
 
 	if(!SQL_SUCCEEDED(r)) {
 		report_error(SQL_HANDLE_STMT, st, r, _("Failed to list tables"));
+		SQLFreeHandle(SQL_HANDLE_STMT, st);
+		res_free(res);
 		return 0;
+	} else if(r == SQL_SUCCESS_WITH_INFO) {
+		fetch_warnings(res, SQL_HANDLE_STMT, st);
 	}
 
-	return fetch_results(st, taken);
+	fetch_results(res, st);
+	res_stop_timer(res);
+
+	return res;
 }
 
-results *get_columns(SQLHDBC conn, const char *catalog,
+results *get_columns(const char *catalog,
 			const char *schema, const char *table)
 {
 	SQLHSTMT st;
+	results *res;
 	SQLRETURN r;
-	struct timeval taken;
 
 	r = SQLAllocHandle(SQL_HANDLE_STMT, conn, &st);
 	if(!SQL_SUCCEEDED(r)) {
@@ -538,36 +613,33 @@ results *get_columns(SQLHDBC conn, const char *catalog,
 		return 0;
 	}
 
-	gettimeofday(&taken, 0);
-	current_statement = st;
+	set_current_statement(&st);
+
+	res = res_alloc();
+	res_start_timer(res);
 
 	r = SQLColumns(st,
 		      (SQLCHAR *) catalog, SQL_NTS,
 		      (SQLCHAR *) schema, SQL_NTS,
 		      (SQLCHAR *) table, SQL_NTS,
 		      (SQLCHAR *) "%", SQL_NTS);
-	time_taken(&taken);
 
 	if(!SQL_SUCCEEDED(r)) {
 		report_error(SQL_HANDLE_STMT, st, r, _("Failed to list columns"));
+		SQLFreeHandle(SQL_HANDLE_STMT, st);
+		res_free(res);
 		return 0;
+	} else if(r == SQL_SUCCESS_WITH_INFO) {
+		fetch_warnings(res, SQL_HANDLE_STMT, st);
 	}
 
-	return fetch_results(st, taken);
-}
+	fetch_results(res, st);
+	res_stop_timer(res);
 
-results *db_list_schemas(SQLHDBC conn, const char *catalog)
-{
-	char *c;
-	results *res;
-
-	c = catalog ? (char *) catalog : get_current_catalog(conn);
-	res = get_tables(conn, c, SQL_ALL_SCHEMAS, 0);
-	if(!catalog) free(c);
 	return res;
 }
 
-results *db_list_tables(SQLHDBC conn, const char *spec)
+results *db_list_tables(const char *spec)
 {
 	char *catalog, *schema;
 	char *buf;
@@ -580,19 +652,19 @@ results *db_list_tables(SQLHDBC conn, const char *spec)
 	if(db_supports_catalogs(conn)) {
 		if(spec) {
 			if(!(buf = strdup(spec))) err_system();
-			parse_catalog_spec(conn, buf, &catalog, &schema);
+			parse_catalog_spec(buf, &catalog, &schema);
 		} else {
-			buf = get_current_catalog(conn);
+			buf = get_current_catalog();
 			catalog = buf;
 		}
 	} else schema = (char *) spec;
 
-	res = get_tables(conn, catalog, schema, 0);
+	res = get_tables(catalog, schema, 0);
 	if(buf) free(buf);
 	return res;
 }
 
-results *db_list_columns(SQLHDBC conn, const char *spec)
+results *db_list_columns(const char *spec)
 {
 	char *catalog, *schema, *table;
 	char *dspec, *buf;
@@ -603,7 +675,7 @@ results *db_list_columns(SQLHDBC conn, const char *spec)
 
 	if(db_supports_catalogs(conn)) {
 
-		parse_catalog_spec(conn, dspec, &catalog, &table);
+		parse_catalog_spec(dspec, &catalog, &table);
 
 		if(table) parse_qualified_table(table, &schema, &table);
 		else {
@@ -616,13 +688,13 @@ results *db_list_columns(SQLHDBC conn, const char *spec)
 		parse_qualified_table(dspec, &schema, &table);
 	}
 
-	res = get_columns(conn, catalog, schema, table);
+	res = get_columns(catalog, schema, table);
 	free(dspec);
 	if(buf) free(buf);
 	return res;
 }
 
-static char *get_current_catalog(SQLHDBC conn)
+static char *get_current_catalog()
 {
 	SQLSMALLINT buflen;
 	char *buf;
@@ -648,8 +720,7 @@ static char *get_current_catalog(SQLHDBC conn)
 	return buf;
 }
 
-static void parse_catalog_spec(SQLHDBC conn, char *spec,
-			     char **catalog, char **schema)
+static void parse_catalog_spec(char *spec, char **catalog, char **schema)
 {
 	SQLUSMALLINT catloc;
 	char sep[8], *p, *q;
@@ -704,12 +775,52 @@ static void parse_qualified_table(char *s, char **schema, char **table)
 	}
 }
 
-static void time_taken(struct timeval *time_taken)
+results *db_autocommit(int change)
 {
-	struct timeval end_time;
+	results *res;
+	SQLUINTEGER state;
+	const char *text;
+	SQLRETURN r;
 
-	gettimeofday(&end_time, 0);
+	res = res_alloc();
+	res_set_cols(res, 1, _("Autocommit state"));
 
-	time_taken->tv_sec  = end_time.tv_sec  - time_taken->tv_sec;
-	time_taken->tv_usec = end_time.tv_usec - time_taken->tv_usec;
+	if(change) {
+		state = change > 0 ? SQL_AUTOCOMMIT_ON : SQL_AUTOCOMMIT_OFF;
+		r = SQLSetConnectAttr(conn, SQL_ATTR_AUTOCOMMIT, (SQLPOINTER) state, 0);
+		if(r == SQL_SUCCESS_WITH_INFO || r == SQL_ERROR)
+			fetch_warnings(res, SQL_HANDLE_DBC, conn);
+	}
+
+	r = SQLGetConnectAttr(conn, SQL_ATTR_AUTOCOMMIT, &state, 0, 0);
+	if(r == SQL_SUCCESS_WITH_INFO || r == SQL_ERROR)
+		fetch_warnings(res, SQL_HANDLE_DBC, conn);
+
+	if(SQL_SUCCEEDED(r)) text = (state == SQL_AUTOCOMMIT_ON) ? _("On") : _("Off");
+	else text = _("(unknown)");
+
+	res_add_row(res, text);
+
+	return res;
 }
+
+results *db_endtran(int commit)
+{
+	results *res;
+	SQLRETURN r;
+
+	res = res_alloc();
+	res_set_nrows(res, -1);
+
+	r = SQLEndTran(SQL_HANDLE_DBC, conn, commit ? SQL_COMMIT : SQL_ROLLBACK);
+	if(!SQL_SUCCEEDED(r)) {
+		report_error(SQL_HANDLE_DBC, conn, r, _("Failed to execute statement"));
+		res_free(res);
+		return 0;
+	} else if(r == SQL_SUCCESS_WITH_INFO) {
+		fetch_warnings(res, SQL_HANDLE_DBC, conn);
+	}
+
+	return res;
+}
+
