@@ -6,10 +6,15 @@
 "\n" \
 "Schema commands:\n" \
 "  catalogs\n" \
-"  schemas [<catalog>]\n" \
+"  schemas\n" \
 "  tables [<catalog>]\n" \
 "  columns <table>\n" \
 "\n" \
+"Transaction commands:\n" \
+"  autocommit on|off\n" \
+"  commit\n" \
+"  rollback\n" \
+"\n" \
 "Other commands:\n" \
 "  set [<variable>] [<value>]\n" \
 "  unset <variable>\n" \
