@@ -29,17 +29,17 @@
 #include "action.h"
 #include "buffer.h"
 #include "db.h"
+#include "output.h"
 #include "parser.h"
 #include "prompt.h"
 #include "rc.h"
 #include "rl.h"
 #include "sig.h"
+#include "stream.h"
 
 
-#include <wchar.h>
-
-const char *dsn, *user, *pass;
-SQLHDBC conn;
+const char *dsn, *user;
+char *pass;
 buffer *mainbuf, *prevbuf;
 
 
@@ -49,6 +49,9 @@ void usage(const char *cmd)
 	       cmd, cmd);
 }
 
+
+#define SWAP_BUFFERS do { tempbuf = mainbuf; mainbuf = prevbuf; prevbuf = tempbuf; } while(0)
+
 int process_line(char *line)
 {
 	char *actionchars, action, *paramstring;
@@ -71,27 +74,12 @@ int process_line(char *line)
 		if(action) {
 			if(action == 'q') return -1;
 
-			if(!mainbuf->next && action != 'r') {
-				if(prevbuf->next) {
-					tempbuf = mainbuf;
-					mainbuf = prevbuf;
-					prevbuf = tempbuf;
-				}
-			}
+			if(!mainbuf->next && prevbuf->next && action != 'r') SWAP_BUFFERS;
 
 			if(action != 'c') {
-				run_action(conn, mainbuf,
-					   action, paramstring);
-
-				if(action == 'e' || action == 'p') {
-					rl_history_add(mainbuf, "");
-				} else {
-					rl_history_add(mainbuf, line - 1);
-				}
-
-				tempbuf = prevbuf;
-				prevbuf = mainbuf;
-				mainbuf = tempbuf;
+				run_action(mainbuf, action, paramstring);
+				rl_history_add(mainbuf, (action == 'e' || action == 'p') ? "" : line -1);
+				SWAP_BUFFERS;
 			}
 
 			mainbuf->next = 0;
@@ -105,11 +93,9 @@ int process_line(char *line)
 		// do nothing
 		break;
 	case BUFFER_COMMAND:
-		run_action(conn, mainbuf, 1, "");
+		run_action(mainbuf, 1, "");
 		rl_history_add(mainbuf, "");
-		tempbuf = prevbuf;
-		prevbuf = mainbuf;
-		mainbuf = tempbuf;
+		SWAP_BUFFERS;
 		mainbuf->next = 0;
 		break;
 	default:
@@ -123,6 +109,7 @@ int main(int argc, char *argv[])
 {
 	int opt;
 	char *line, *p;
+	results *r;
 
 	setlocale(LC_ALL, "");
 
@@ -134,7 +121,8 @@ int main(int argc, char *argv[])
 	while((opt = getopt(argc, argv, "lv")) != -1) {
 		switch(opt) {
 		case 'l':
-			list_all_dsns();
+			r = db_drivers_and_dsns();
+			output_results(r, 1, stream_create(stdout));
 			return 0;
 			break;
 		case 'v':
@@ -165,7 +153,7 @@ int main(int argc, char *argv[])
 		for(p = argv[optind]; *p; p++) *p = 'x';
 	}
 
-	if(!(conn = db_connect())) exit(1);
+	if(!db_connect()) exit(1);
 
 	rl_history_start();
 	signal_handlers_install();
@@ -175,7 +163,7 @@ int main(int argc, char *argv[])
 
 	/* Main loop */
 	for(;;) {
-		line = rl_readline(prompt_render(conn, mainbuf));
+		line = rl_readline(prompt_render(mainbuf));
 		if(!line || process_line(line)) {
 			fputc('\n', stdout);
 			break;
@@ -183,10 +171,14 @@ int main(int argc, char *argv[])
 		free(line);
 	}
 
+	buffer_free(mainbuf);
+	buffer_free(prevbuf);
+
 	rl_history_end();
 
-	SQLDisconnect(conn);
-	SQLFreeHandle(SQL_HANDLE_DBC, conn);
+	db_close();
+
+	if(pass) free(pass);
 
 	return 0;
 }
